@@ -1,4 +1,3 @@
-#!/usr/bin/env python
 # coding: utf-8
 
 # FIXME python2
@@ -11,6 +10,7 @@ import os
 import re
 import uuid
 import zlib
+from copy import deepcopy
 
 from construct import Container, ChecksumError
 from lxml import etree
@@ -26,15 +26,44 @@ from pykeepass.xpath import attachment_xp, entry_xp, group_xp, path_xp
 
 logger = logging.getLogger(__name__)
 
+
+BLANK_DATABASE_FILENAME = "blank_database.kdbx"
+BLANK_DATABASE_LOCATION = os.path.join(os.path.dirname(os.path.realpath(__file__)), BLANK_DATABASE_FILENAME)
+BLANK_DATABASE_PASSWORD = "password"
+
+
 # FIXME python2
 @python_2_unicode_compatible
 class PyKeePass(object):
+    """Open a KeePass database
+
+    Args:
+        filename (:obj:`str`, optional): path to database.  If None, the
+            path given when the database was opened is used.
+        password (:obj:`str`, optional): database password.  If None,
+            database is assumed to have no password
+        keyfile (:obj:`str`, optional): path to keyfile.  If None,
+            database is assumed to have no keyfile
+        transformed_key (:obj:`bytes`, optional): precomputed transformed
+            key.
+
+    Raises:
+        CredentialsError: raised when password/keyfile or transformed key
+            are wrong
+        HeaderChecksumError: raised when checksum in database header is
+            is wrong.  e.g. database tampering or file corruption
+        PayloadChecksumError: raised when payload blocks checksum is wrong,
+            e.g. corruption during database saving
+
+    Todo:
+        - raise, no filename provided, database not open
+    """
 
     def __init__(self, filename, password=None, keyfile=None,
                  transformed_key=None):
-        self.filename = filename
 
         self.read(
+            filename=filename,
             password=password,
             keyfile=keyfile,
             transformed_key=transformed_key
@@ -49,9 +78,17 @@ class PyKeePass(object):
 
     def read(self, filename=None, password=None, keyfile=None,
              transformed_key=None):
+        """
+        See class docstring.
+
+        Todo:
+            - raise, no filename provided, database not open
+        """
         self.password = password
         self.keyfile = keyfile
-        if not filename:
+        if filename:
+            self.filename = filename
+        else:
             filename = self.filename
 
         try:
@@ -61,25 +98,46 @@ class PyKeePass(object):
                 keyfile=keyfile,
                 transformed_key=transformed_key
             )
-        except ChecksumError:
-            raise CredentialsIntegrityError
+        except ChecksumError as e:
+            if e.path in (
+                    '(parsing) -> body -> cred_check', # KDBX4
+                    '(parsing) -> cred_check' # KDBX3
+                    ):
+                raise CredentialsError
+            elif e.path == '(parsing) -> body -> sha256':
+                raise HeaderChecksumError
+            elif e.path in (
+                    '(parsing) -> body -> payload -> hmac_hash', # KDBX4
+                    '(parsing) -> xml -> block_hash' # KDBX3
+                    ):
+                raise PayloadChecksumError
+            else:
+                raise
 
     def save(self, filename=None, transformed_key=None):
+        """Save current database object to disk.
+
+        Args:
+            filename (:obj:`str`, optional): path to database.  If None, the
+                path given when the database was opened is used.
+            transformed_key (:obj:`bytes`, optional): precomputed transformed
+                key.
+        """
         if not filename:
             filename = self.filename
 
-        with open(filename, 'wb') as f:
-            f.write(
-                KDBX.build(
-                    self.kdbx,
-                    password=self.password,
-                    keyfile=self.keyfile,
-                    transformed_key=transformed_key
-                )
-            )
+        KDBX.build_file(
+            self.kdbx,
+            filename,
+            password=self.password,
+            keyfile=self.keyfile,
+            transformed_key=transformed_key
+        )
 
     @property
     def version(self):
+        """tuple: Length 2 tuple of ints containing major and minor versions.
+        Generally (3, 1) or (4, 0)."""
         return (
             self.kdbx.header.value.major_version,
             self.kdbx.header.value.minor_version
@@ -87,10 +145,14 @@ class PyKeePass(object):
 
     @property
     def encryption_algorithm(self):
+        """str: encryption algorithm used by database during decryption.
+        Can be one of 'aes256', 'chacha20', or 'twofish'."""
         return self.kdbx.header.value.dynamic_header.cipher_id.data
 
     @property
     def kdf_algorithm(self):
+        """str: key derivation algorithm used by database during decryption.
+        Can be one of 'aeskdf', 'argon2', or 'aeskdf'"""
         if self.version == (3, 1):
             return 'aeskdf'
         elif self.version == (4, 0):
@@ -102,30 +164,52 @@ class PyKeePass(object):
 
     @property
     def transformed_key(self):
+        """bytes: transformed key used in database decryption.  May be cached
+        and passed to `open` for faster database opening"""
         return self.kdbx.body.transformed_key
 
     @property
     def tree(self):
+        """lxml.etree._ElementTree: database XML payload"""
         return self.kdbx.body.payload.xml
 
     @property
     def root_group(self):
+        """Group: root Group of database"""
         return self.find_groups(path='', first=True)
 
     @property
     def groups(self):
-        return self.find_groups_by_name('.*', regex=True)
+        """:obj:`list` of :obj:`Group`: list of all Group objects in database
+        """
+        return self._xpath('//Group', cast=True)
 
     @property
     def entries(self):
-        return self.find_entries_by_title('.*', regex=True)
+        """:obj:`list` of :obj:`Entry`: list of all Entry objects in database,
+        excluding history"""
+        return self._xpath('//Entry', cast=True)
 
-    def dump_xml(self, outfile):
-        '''
-        Dump the content of the database to a file
-        NOTE The file is unencrypted!
-        '''
-        with open(outfile, 'wb') as f:
+    def xml(self):
+        """Get XML part of database as string
+
+        Returns:
+            str: XML payload section of database.
+        """
+        return etree.tostring(
+            self.tree,
+            pretty_print=True,
+            standalone=True,
+            encoding='unicode'
+        )
+
+    def dump_xml(self, filename):
+        """ Dump the contents of the database to file as XML
+
+        Args:
+            filename (str): path to output file
+        """
+        with open(filename, 'wb') as f:
             f.write(
                 etree.tostring(
                     self.tree,
@@ -137,6 +221,26 @@ class PyKeePass(object):
 
     def _xpath(self, xpath_str, tree=None, first=False, history=False,
                cast=False, **kwargs):
+        """Look up elements in the XML payload and return corresponding object.
+
+        Internal function which searches the payload lxml ElementTree for
+        elements via XPath.  Matched entry, group, and attachment elements are
+        automatically cast to their corresponding objects, otherwise an error
+        is raised.
+
+        Args:
+            xpath_str (str): XPath query for finding element(s)
+            tree (:obj:`_ElementTree`, :obj:`Element`, optional): use this
+                element as root node when searching
+            first (bool): If True, function returns first result or None.  If
+                False, function returns list of matches or empty list.  Default
+                is False.
+            history (bool): If True, history entries are included in results.
+                Default is False.
+            cast (bool): If True, matches are instead instantiated as
+                pykeepass Group, Entry, or Attachment objects.  An exception
+                is raised if a match cannot be cast.  Default is False.
+        """
 
         if tree is None:
             tree = self.tree
@@ -207,6 +311,10 @@ class PyKeePass(object):
             if 'uuid' in kwargs.keys():
                 kwargs['uuid'] = base64.b64encode(kwargs['uuid'].bytes).decode('utf-8')
 
+            # convert tags to semicolon separated string before building xpath
+            if 'tags' in kwargs.keys():
+                kwargs['tags'] = ';'.join(kwargs['tags'])
+
             # build xpath to filter results with specified attributes
             for key, value in kwargs.items():
                 if key not in keys_xp[regex].keys():
@@ -476,13 +584,13 @@ class PyKeePass(object):
         else:
             binaries = []
             for elem in self._xpath('/KeePassFile/Meta/Binaries/Binary'):
-                if elem.attrib['Compressed'] == 'True':
+                if elem.get('Compressed') == 'True':
                     data = zlib.decompress(
                         base64.b64decode(elem.text),
                         zlib.MAX_WBITS | 32
                     )
                 else:
-                    data = base64.b64decode(elem.text).decode()
+                    data = base64.b64decode(elem.text)
                 binaries.insert(int(elem.attrib['ID']), data)
 
         return binaries
@@ -514,7 +622,7 @@ class PyKeePass(object):
             data = base64.b64encode(data).decode()
 
             # set ID for Binary Element
-            binary_id = len(self.binaries) - 1
+            binary_id = len(self.binaries)
 
             # add binary element to XML
             binaries.append(
@@ -547,3 +655,15 @@ class PyKeePass(object):
         )
         for reference in binaries_gt:
             reference.id = reference.id - 1
+
+
+def create_database(filename, password=None, keyfile=None, transformed_key=None):
+    keepass_instance = PyKeePass(BLANK_DATABASE_LOCATION, BLANK_DATABASE_PASSWORD)
+
+    keepass_instance.filename = filename
+    keepass_instance.password = password
+    keepass_instance.keyfile = keyfile
+
+    keepass_instance.save(transformed_key)
+
+    return keepass_instance
