@@ -18,7 +18,7 @@ from pykeepass.entry import Entry
 from pykeepass.exceptions import BinaryError
 from pykeepass.group import Group
 from pykeepass.kdbx_parsing import KDBX
-from pykeepass.exceptions import BinaryError, CredentialsIntegrityError
+from pykeepass.exceptions import BinaryError, CredentialsError
 
 """
 Missing Tests:
@@ -45,11 +45,21 @@ class KDBX3Tests(unittest.TestCase):
 
     # get some things ready before testing
     def setUp(self):
+        shutil.copy(
+            os.path.join(base_dir, self.database),
+            os.path.join(base_dir, 'change_creds.kdbx')
+        )
         self.kp = PyKeePass(
             os.path.join(base_dir, self.database),
             password=self.password,
             keyfile=os.path.join(base_dir, self.keyfile)
         )
+        # for tests which modify the database, use this
+        self.kp_tmp = PyKeePass(
+            os.path.join(base_dir, 'change_creds.kdbx'),
+            password=self.password,
+            keyfile=os.path.join(base_dir, self.keyfile)
+        )
 
 
 class KDBX4Tests(KDBX3Tests):
@@ -687,30 +697,15 @@ class AttachmentTests3(KDBX3Tests):
 
 
 class PyKeePassTests3(KDBX3Tests):
-    def setUp(self):
-        shutil.copy(
-            os.path.join(base_dir, self.database),
-            os.path.join(base_dir, 'change_creds.kdbx')
-        )
-        self.kp = PyKeePass(
-            os.path.join(base_dir, self.database),
-            password=self.password,
-            keyfile=os.path.join(base_dir, self.keyfile)
-        )
-        self.kp_tmp = PyKeePass(
-            os.path.join(base_dir, 'change_creds.kdbx'),
-            password=self.password,
-            keyfile=os.path.join(base_dir, self.keyfile)
-        )
 
     def test_set_credentials(self):
         self.kp_tmp.password = 'f00bar'
         self.kp_tmp.keyfile = os.path.join(base_dir, 'change.key')
         self.kp_tmp.save()
         self.kp_tmp = PyKeePass(
-            os.path.join(base_dir, 'change_creds.kdbx'),
-            password='f00bar',
-            keyfile=os.path.join(base_dir, 'change.key')
+            self.kp_tmp.filename,
+            'f00bar',
+            self.kp_tmp.keyfile
         )
 
         results = self.kp.find_entries_by_username('foobar_user', first=True)
@@ -727,20 +722,35 @@ class PyKeePassTests3(KDBX3Tests):
 
 
 class BugRegressionTests3(KDBX3Tests):
-    def test_129(self):
+    def test_issue129(self):
         # issue 129 - protected multiline string fields lose newline
         e = self.kp.find_entries(title='foobar_entry', first=True)
         self.assertEqual(e.get_custom_property('multiline'), 'hello\nworld')
 
     def test_pull102(self):
         # PR 102 - entries are protected after save
-        # reset self.kp
+        # reset self.kp_tmp
         self.setUp()
-        e = self.kp.find_entries(title='foobar_entry', first=True)
+        e = self.kp_tmp.find_entries(title='foobar_entry', first=True)
         self.assertEqual(e.password, 'foobar')
-        self.kp.save()
+        self.kp_tmp.save()
         self.assertEqual(e.password, 'foobar')
 
+    def test_issue193(self):
+        # issue 193 - kp.entries doesn't return entries with title=None
+        e = self.kp.add_entry(self.kp.root_group, 'test', 'user', 'pass')
+        prop = e._xpath('String/Key[text()="Title"]/..', first=True)
+        e._element.remove(prop)
+        self.assertTrue(e.title is None)
+        self.assertTrue(e in self.kp.entries)
+        # also test for kp.groups
+        g = self.kp.add_group(self.kp.root_group, 'test_g')
+        prop = g._xpath('Name', first=True)
+        g._element.remove(prop)
+        self.assertTrue(g.name is None)
+        self.assertTrue(g in self.kp.groups)
+
+
 
 class EntryFindTests4(KDBX4Tests, EntryFindTests3):
     pass
@@ -882,7 +892,7 @@ class KDBXTests(unittest.TestCase):
             )
 
     def test_open_error(self):
-        with self.assertRaises(CredentialsIntegrityError):
+        with self.assertRaises(CredentialsError):
             database = 'test4.kdbx'
             invalid_password = 'foobar'
             keyfile = os.path.join(base_dir, 'test4.key')
@@ -891,7 +901,7 @@ class KDBXTests(unittest.TestCase):
                 password=invalid_password,
                 keyfile=keyfile
             )
-        with self.assertRaises(CredentialsIntegrityError):
+        with self.assertRaises(CredentialsError):
             database = 'test4.kdbx'
             password = 'password'
             invalid_keyfile = os.path.join(base_dir, 'test3.key')
